/*
 * This file is part of the Chelsio T3 Ethernet driver.
 *
 * Copyright (C) 2003-2007 Chelsio Communications.  All rights reserved.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the LICENSE file included in this
 * release for licensing terms and conditions.
 */
#ifndef __CXGB3_COMPAT_H
#define __CXGB3_COMPAT_H

#include <linux/version.h>
#include "common.h"
#include <linux/pci.h>

/* XXX Verify OS version */
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,13) && \
    LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,5)

#if LINUX_VERSION_CODE == KERNEL_VERSION(2,6,5)

struct msix_entry {
	u16 	vector;	/* kernel uses to write allocated vector */
	u16	entry;	/* driver uses to specify entry, OS writes */
};

static inline void pci_disable_msi(struct pci_dev *dev)
{}

static inline int pci_enable_msix(struct pci_dev* dev, struct msix_entry *entries,
			          int nvec)
{
	return -1;
}

static inline void pci_disable_msix(struct pci_dev* dev)
{}

static inline struct mii_ioctl_data *if_mii(struct ifreq *rq)

{
	return (struct mii_ioctl_data *) &rq->ifr_ifru;
}

#define _spin_trylock spin_trylock

#endif /* KERNEL_VERSION(2.6.5) */

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,11)
static inline int t3_os_pci_save_state(struct adapter *adapter)
{
	return pci_save_state(adapter->pdev, adapter->t3_config_space);
}

static inline int t3_os_pci_restore_state(struct adapter *adapter)
{
	return pci_restore_state(adapter->pdev, adapter->t3_config_space);
}

static
inline void cancel_rearming_delayed_workqueue(struct workqueue_struct *wq,
					      struct work_struct *work)
{
	while (!cancel_delayed_work(work))
		flush_workqueue(wq);
}

#else
static inline int t3_os_pci_save_state(adapter_t *adapter)
{
	return pci_save_state(adapter->pdev);
}

static inline int t3_os_pci_restore_state(adapter_t *adapter)
{
	return pci_restore_state(adapter->pdev);
}
#endif

static inline int __netif_rx_schedule_prep(struct net_device *dev)
{
	return !test_and_set_bit(__LINK_STATE_RX_SCHED, &dev->state);
}

#ifndef CONFIG_DEBUG_FS
#include <linux/err.h>
/* Adapted from debugfs.h */
static inline struct dentry *debugfs_create_dir(const char *name,
						struct dentry *parent)
{
	return ERR_PTR(-ENODEV);
}

static inline void debugfs_remove(struct dentry *dentry)
{}
#else
#include <linux/debugfs.h>
#endif

static inline void setup_timer(struct timer_list * timer,
				void (*function)(unsigned long),
				unsigned long data)
{
	timer->function = function;
	timer->data = data;
	init_timer(timer);
}

#define DEFINE_MUTEX DECLARE_MUTEX
#define mutex_lock down
#define mutex_unlock up

#undef DEFINE_RWLOCK /* broken RH4u3 definition, rw_lock_t does not exist */
#define DEFINE_RWLOCK(x)	rwlock_t x = RW_LOCK_UNLOCKED

#define gfp_t unsigned

/* 2.6.14 and above */
#elif LINUX_VERSION_CODE > KERNEL_VERSION(2,6,13)
#include <linux/debugfs.h>

static inline int t3_os_pci_save_state(adapter_t *adapter)
{
	return pci_save_state(adapter->pdev);
}

static inline int t3_os_pci_restore_state(adapter_t *adapter)
{
	return pci_restore_state(adapter->pdev);
}

#endif /* LINUX_VERSION_CODE */

extern int netdev_nit;


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)

typedef irqreturn_t (*intr_handler_t)(int, void *, struct pt_regs *);
#define DECLARE_INTR_HANDLER(handler, irq, cookie, regs) \
	static irqreturn_t handler(int irq, void *cookie, struct pt_regs *regs)

intr_handler_t t3_intr_handler(struct adapter *adap, int polling);
static inline void t3_poll_handler(struct adapter *adapter,
				   struct sge_qset *qs)
{
	t3_intr_handler(adapter, qs->rspq.polling) (0,
		(adapter->flags & USING_MSIX) ? (void *)qs : (void *)adapter,
		NULL);
}

#define CHECKSUM_PARTIAL CHECKSUM_HW
#define CHECKSUM_COMPLETE CHECKSUM_HW

#ifndef I_PRIVATE
#define i_private u.generic_ip
#endif

#else /* 2.6.19 */
typedef irqreturn_t (*intr_handler_t)(int, void *);
#define DECLARE_INTR_HANDLER(handler, irq, cookie, regs) \
	static irqreturn_t handler(int irq, void *cookie)

intr_handler_t t3_intr_handler(struct adapter *adap, int polling);
static inline void t3_poll_handler(struct adapter *adapter,
		 		   struct sge_qset *qs)
{
	t3_intr_handler(adapter, qs->rspq.polling) (0,
		(adapter->flags & USING_MSIX) ? (void *)qs : (void *)adapter);
}

#endif /* 2.6.19 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
#define DECLARE_TASK_FUNC(task, task_param) \
	static void task(void *task_param)

#define WORK2ADAP(task_param, task) task_param
#define DELWORK2ADAP(task_param, task) task_param
#define WORK2T3CDATA(task_param, task) task_param

#define delayed_work work_struct

#define T3_INIT_WORK INIT_WORK
#define T3_INIT_DELAYED_WORK INIT_WORK

#else /* 2.6.20 */

#define DECLARE_TASK_FUNC(task, task_param) \
	static void task(struct work_struct *task_param)

#define WORK2ADAP(task_param, task) \
	container_of(task_param, struct adapter, task)

#define DELWORK2ADAP(task_param, task) \
	container_of(task_param, struct adapter, task.work)

#define WORK2T3CDATA(task_param, task) \
	container_of(task_param, struct t3c_data, task)

#define T3_INIT_WORK(task_handler, task, adapter) \
	INIT_WORK(task_handler, task)

#define T3_INIT_DELAYED_WORK(task_handler, task, adapter) \
	INIT_DELAYED_WORK(task_handler, task)

#endif /* 2.6.20 */

#if defined(CONFIG_FW_LOADER) || defined(CONFIG_FW_LOADER_MODULE)
#include <linux/firmware.h>
#else
struct firmware {
	size_t size;
	u8 *data;
};

struct device;

static inline int request_firmware(const struct firmware **firmware_p,
				   char *name,
				   struct device *device)
{
	printk(KERN_WARNING
	       "FW_LOADER not set in this kernel. FW upgrade aborted.\n");
	return -1;
}

static inline void release_firmware(const struct firmware *fw)
{}
#endif /* FW_LOADER */

#if !defined(RTNL_TRYLOCK)
#include <linux/rtnetlink.h>
static inline int rtnl_trylock(void)
{
	return !rtnl_shlock_nowait();
}
#endif /* RTNL_TRYLOCK */

#ifndef GSO_SIZE
#define gso_size tso_size
#endif /* GSO_SIZE */

#ifndef NIPQUAD_FMT
#define NIPQUAD_FMT "%u.%u.%u.%u"
#endif

/* sysfs compatibility */
#if  LINUX_VERSION_CODE < KERNEL_VERSION(2,6,21)

#define to_net_dev(class) container_of(class, struct net_device, class_dev)

#define cxgb3_compat_device class_device

#define CXGB3_SHOW_FUNC(func, d, attr, buf)			\
	static ssize_t func(struct cxgb3_compat_device *d,	\
			    char *buf)				\

#define CXGB3_STORE_FUNC(func, d, attr, buf, len)		\
	static ssize_t func(struct cxgb3_compat_device *d,	\
			    const char *buf,			\
			    size_t len)

#define CXGB3_DEVICE_ATTR(_name,_mode,_show,_store)		\
struct class_device_attribute dev_attr_##_name = 		\
	__ATTR(_name,_mode,_show,_store)

static inline struct kobject *net2kobj(struct net_device *dev)
{
	return &dev->class_dev.kobj;
}

#else /* sysfs compatibility */

#define cxgb3_compat_device device

#define CXGB3_SHOW_FUNC(func, d, attr, buf)			\
	static ssize_t func(struct cxgb3_compat_device *d,	\
			    struct device_attribute *attr,	\
			    char *buf)				\

#define CXGB3_STORE_FUNC(func, d, attr, buf, len)		\
	static ssize_t func(struct cxgb3_compat_device *d,	\
			    struct device_attribute *attr,	\
			    const char *buf,			\
			    size_t len)

#define CXGB3_DEVICE_ATTR DEVICE_ATTR

static inline struct kobject *net2kobj(struct net_device *dev)
{
	return &dev->dev.kobj;
}

#endif /* sysfs compatibility */

#if !defined(IRQF)
#define IRQF_SHARED SA_SHIRQ
#endif /* IRQF */

#if !defined(VLANGRP)
#include <linux/if_vlan.h>
static inline struct net_device *vlan_group_get_device(struct vlan_group *vg,
						       int vlan_id)
{
	return vg->vlan_devices[vlan_id];
}
#endif /* VLANGRP */

#if !defined(for_each_netdev)
#define for_each_netdev(d) \
	for (d = dev_base; d; d = d->next)
#endif

#include <linux/ip.h>

#if !defined(NEW_SKB_COPY)
static inline void skb_copy_from_linear_data(const struct sk_buff *skb,
					     void *to,
					     const unsigned int len)
{
	memcpy(to, skb->data, len);
}

static inline void skb_copy_to_linear_data(struct sk_buff *skb,
					   const void *from,
					   const unsigned int len)
{
	memcpy(skb->data, from, len);
}
#endif

#if defined(NEW_SKB_OFFSET)
static inline void cxgb3_set_skb_header(struct sk_buff *skb,
					struct iphdr *ip_hdr,
					int offset)
{
	skb_set_network_header(skb, offset);
}

#else /* NEW_SKB_OFFSET */
static inline int skb_network_offset(struct sk_buff *skb)
{
	return skb->nh.raw - skb->data;
}

static inline unsigned char *skb_transport_header(const struct sk_buff *skb)
{
	return skb->h.raw;
}

static inline int skb_transport_offset(const struct sk_buff *skb)
{
	return skb->h.raw - skb->data;
}

#if !defined(T3_IP_HDR)
static inline struct iphdr *ip_hdr(const struct sk_buff *skb)
{
	return skb->nh.iph;
}
#endif

static inline struct tcphdr *tcp_hdr(const struct sk_buff *skb)
{
	return skb->h.th;
}

static inline void skb_reset_mac_header(struct sk_buff *skb)
{
	skb->mac.raw = skb->data;
}

static inline void skb_reset_network_header(struct sk_buff *skb)
{
	skb->nh.raw = skb->data;
}

static inline void skb_reset_transport_header(struct sk_buff *skb)
{
	skb->h.raw = skb->data;
}

static inline void cxgb3_set_skb_header(struct sk_buff *skb,
					struct iphdr *ip_hdr,
					int offset)
{
	skb->nh.iph = ip_hdr;
}

#endif /* NEW_SKB_OFFSET */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
#if defined(ETHTOOL_GPERMADDR)
#define CXGB3_ETHTOOL_GPERMADDR ETHTOOL_GPERMADDR
#endif
#endif

#if !defined(TRANSPORT_HEADER)
#define transport_header h.raw
#endif
#endif
